#ifndef HAL_ISRDISPATCH_CPP
#define HAL_ISRDISPATCH_CPP

//----------------------------------
// Includes and forward declarations
//----------------------------------
#include "cyg/hal/hal_isrdispatch.h"

#if !defined(CYG_HAL_STARTUP_STUBS) && defined(CYG_HAL_MPC555)
#include "cyg/hal/hal_intr.h"

//------------------------
// Definition of constants
//------------------------
#define Cyg_InterruptHANDLED  1
#define Cyg_InterruptCALL_DSR 2

//-------------------------------
// External function declarations
//-------------------------------
externC void cyg_interrupt_post_dsr(CYG_ADDRWORD interrupt_object);

//---------------------
// Class implementation
//---------------------
// Public constructor(s) and destructor
HAL_isrDispatcher::HAL_isrDispatcher()
     : m_vector(0)
{ // Masking and unmasking with this will have no effect. This is what we want before the 
} // class is initialised

// Provide an (empty) virtual destructor
HAL_isrDispatcher::~HAL_isrDispatcher()
{
}

//------------------------------
// Public members implementation
//------------------------------
// enable
// After calling this function, the dispatcher will effectively
// see the interrupts that are meant for him
// This does not mean that the interrupt is enabled, it only means that if the
// interrupt occurs, the dispatcher will see it.
void HAL_isrDispatcher::enableIsrDispatching(void)
{ // This is safe, or the vector is 0, and cannot be masked, or it is valid
  if(m_vector)
    HAL_INTERRUPT_UNMASK(m_vector);
}

// disable
// after calling this function, the distpatcher will no longer see
// any interrupts. This does not disable the interrupts
void HAL_isrDispatcher::disableIsrDispatching(void)
{ // This is sdafe, or the vector is 0, and cannot be unmasked, or it is valid
  if(m_vector)
    HAL_INTERRUPT_MASK(m_vector);
}

// enableInterrupts
void HAL_isrDispatcher::enableInterrupts(void)
{
  enableInterruptingDevice();
}

// disableInterrupts
void HAL_isrDispatcher::disableInterrupts(void)
{
  disableInterruptingDevice();
}

// attachToLevel
bool HAL_isrDispatcher::attach(unsigned char a_level)
{
  if(m_vector)  // If already active
    detach();

  CYG_ASSERT(a_level < 8, "Dispatcher level 0..7");
  m_vector = (a_level + 1) * 2; // level 0->vector 2, level 1->vector 4 ....

  // Make sure interrupts do not start prior to being asked
  disableIsrDispatching();
  if(setInterruptingLevel(a_level))
  {  
    HAL_INTERRUPT_ATTACH(m_vector, sDispatch, this, 0);
    return true;
  }
  else
    return false;
}

// detach
void HAL_isrDispatcher::detach(void)
{
  if(m_vector)
  { // No use in generating additional interrupts if no one sees them
    disableIsrDispatching();
    HAL_INTERRUPT_DETACH(m_vector, sDispatch);
    m_vector = 0;
  }
}

//-------------------------
// Private member functions
//-------------------------
// sDispatch
cyg_uint32 HAL_isrDispatcher::sDispatch(CYG_ADDRWORD a_vector, HAL_isrDispatcher * a_dispatcher)
{
  CYG_ASSERT(a_dispatcher, "Dispatcher is a NULL pointer");

  HAL_INTERRUPT_MASK(a_vector);

  cyg_uint32 result = a_dispatcher->dispatch(a_vector);
  CYG_ASSERT(result & Cyg_InterruptHANDLED, "Interrupt not handled");

  HAL_INTERRUPT_UNMASK(a_vector);

  return result;
}

// dispatch
cyg_uint32 HAL_isrDispatcher::dispatch(CYG_ADDRWORD a_vector)
{
  CYG_ASSERT(a_vector == m_vector, "Corrupted interrupt system");

  HAL_INTERRUPT_ACKNOWLEDGE(a_vector);
  cyg_uint32 vector = findInterruptingVector();
  
  if(vector)  
    return callISR(vector);
  else
    return 0;
}

// callISR
cyg_uint32 HAL_isrDispatcher::callISR(CYG_ADDRWORD a_vector)
{
  isr theIsr = (isr)(hal_interrupt_handlers[a_vector]);
  CYG_ADDRWORD data =   hal_interrupt_data[a_vector];
  cyg_uint32 result = (theIsr)(a_vector, data);

#ifdef CYGFUN_HAL_COMMON_KERNEL_SUPPORT
  if(result & Cyg_InterruptCALL_DSR)
    cyg_interrupt_post_dsr(hal_interrupt_objects[a_vector]);
#endif

  // Mask out everything except the handled flag
  return result & Cyg_InterruptHANDLED;
}
#endif // if !defined ...

#endif // HAL_ISRDISPATCH_CPP
