##===========================================================================
##
##                                Vectors.s
##                        MPC555 exception vectors
##
##===========================================================================

##---------------------------------------------------------------------------
## Some preprocessor stuff
#include <pkgconf/hal.h>
#include <pkgconf/system.h>

#ifdef CYGPKG_KERNEL
#include <pkgconf/kernel.h>                    // CYGPKG_KERNEL_INSTRUMENT
#endif 

#define CYGARC_HAL_COMMON_EXPORT_CPU_MACROS    // Make short names available
#include "cyg/hal/mpc555_regs.h"
#include "cyg/hal/mpc555.inc"
##
##---------------------------------------------------------------------------

	.file "vectors.s"

	.extern hal_interrupt_handlers
	.extern hal_interrupt_data
	.extern hal_interrupt_objects
	.extern cyg_instrument
	.extern hal_hardware_init
        .global __default_interrupt_vsr
	.global __default_exception_vsr
        .global restore_state

##===========================================================================
## MSR initialization
## zero out MSR except
## FP - floating point available
## ME - machine check enable
## RI - recoverable interrupt

#define CYG_MSR_COMMON (MSR_FP | MSR_ME | MSR_RI)
#if defined(CYG_HAL_STARTUP_ROM) || defined(CYG_HAL_STARTUP_STUBS)
#define CYG_MSR CYG_MSR_COMMON
#else
#define CYG_MSR CYG_MSR_COMMON
#endif

##============================================================================
## The powerPC EABI defines registers R13..R31 as callee saved and thus
## we do not need them when calling C-functions
#ifdef CYGDBG_HAL_COMMON_INTERRUPTS_SAVE_MINIMUM_CONTEXT
#define MAX_SAVE_REG  13
#define MAX_SAVE_FREG 13
#else
#define MAX_SAVE_REG  31
#define MAX_SAVE_FREG 31
#endif

	.section ".vectors", "ax"
	.global  __reset_vector
	.extern  __start

##============================================================================
## Define the exception vector macro

	.macro exception_vector name
	.p2align 8                      ## advance section counter till last 8 bits are zero 
	.global __exception_\name
__exception_\name:
	mtspr sprg1, r3			## stash some work registers away
	mtspr sprg2, r4			
	mtspr sprg3, r5
	mfcr r4				## stash CR
	li r5, __exception_\name@L	## load low half of vector address
	srwi r5, r5, 6			## shift right by 6
	lwi r3, hal_vsr_table		## table base
	lwzx r3, r3, r5			## address of vsr
	mflr r5				## save link register
	mtlr r3				## store vsr in it
	li r3, __exception_\name@L	## reload low half of vector address
	blr				## go to common code
	.endm

##============================================================================
## Now define the exception vector table
	.org 0x00			## set the section counter to zero

        ## Now fill in the interrupt vector table
        ## Only the reset vector will be treated differently
        ## All other interrupts will call the interrupthandler with the
        ## interruptnumber

#if defined(CYG_HAL_STARTUP_ROM) || defined(CYG_HAL_STARTUP_STUBS)
 	.org	0x100
#endif

__reset_vector:
#if defined(CYG_HAL_STARTUP_RAM)
	b	__start
#else
	b	(__start+0x400000)	## use this for booting starting from code in flash
#endif

#if defined(CYG_HAL_STARTUP_ROM) || defined(CYG_HAL_STARTUP_STUBS)
	## We do not need this table if we are starting from RAM, we will just make sure that the
	## VSR tables are accessible by both the monitor and the application and everything works fine
	exception_vector	machine_check			## 0x00200
	exception_vector	reserved_00300			## 0x00300
	exception_vector	reserved_00400			## 0x00400
	exception_vector	external			## 0x00500
	exception_vector	alignment			## 0x00600
	exception_vector	program				## 0x00700
	exception_vector	floatingpoint_unavailable	## 0x00800
	exception_vector	decrementer			## 0x00900
	exception_vector	reserved_00a00			## 0x00a00
	exception_vector	reserved_00b00			## 0x00b00
	exception_vector	system_call			## 0x00c00
	exception_vector	trace				## 0x00d00
	exception_vector	floatingpoint_assist		## 0x00e00
	exception_vector	reserved_00f00			## 0x00f00
	exception_vector	software_emu			## 0x01000
	exception_vector	reserved_001100			## 0x01100
	exception_vector	reserved_001200			## 0x01200
	exception_vector	instruction_tlb_error		## 0x01300
	exception_vector	data_tlb_error			## 0x01400
	exception_vector	reserved_01500			## 0x01500
	exception_vector	reserved_01600			## 0x01600
	exception_vector	reserved_01700			## 0x01700
	exception_vector	reserved_01800			## 0x01800
	exception_vector	reserved_01900			## 0x01900
	exception_vector	reserved_01a00			## 0x01a00
	exception_vector	reserved_01b00			## 0x01b00
	exception_vector	data_breakpoint			## 0x01c00
	exception_vector	instruction_breakpoint		## 0x01d00
	exception_vector	peripheral_breakpoint		## 0x01e00
	exception_vector	NMI_port			## 0x01f00
#endif

	##-----------------------------------------------------------------
	## This code handles the common part of all exception handlers.
	## It saves the machine state onto the stack and then calls a 'C'		
	## routine to do the rest of the work. This may result in thread
	## switches and changes to the saved state. When we return here, the
	## saved state is restored and execution is continued.
	.text
__default_exception_vsr:
	## We come here with all registers containing their pre-execution values
	## except :
	## R3	 = least 16 bits of vector address
	## R4	 = saved CR
	## R5	 = saved LR
	## LR	 = VSR address
	## SPRG1 = old R3
	## SPRG2 = old R4
	## SPRG3 = old R5
	## SRR0	 = old PC
	## SRR1	 = old MSR and the exception cause (the POW state is lost)
	subi r1, r1, ppc_exception_decrement
					## Leave space for registers and a safety margin
	## First, save away some registers
	stw r3, ex_ppcreg_vector(r1)	## stash vector
	stw r4, ex_ppcreg_cr(r1)	## stash CR
	stw r5, ex_ppcreg_lr(r1)	## stash LR

	## Enable MMU
	lwi r3, CYG_MSR
	sync
	mtmsr r3
	sync

	mfspr r3, sprg1			## save original r3
	stw r3, ex_ppcreg_reg+3*4(r1)
	mfspr r4, sprg2			## save original r4
	stw r4, ex_ppcreg_reg+4*4(r1)
	mfspr r5, sprg3			## save original r5
	stw r5, ex_ppcreg_reg+5*4(r1)

	stw r0, ex_ppcreg_reg+0*4(r1)	## save r0
	stw r2, ex_ppcreg_reg+2*4(r1)	## save r2

	mr r3, r1			## recreate original r1 and save it in state
	addi r3, r3, ppc_exception_decrement
	stw r3, ex_ppcreg_reg+1*4(r1)

	## Now save all other integer registers
	.set _reg, 6
	.rept MAX_SAVE_REG+1-6
	stw _reg, ex_ppcreg_reg+_reg*4(r1)
	.set _reg, _reg+1
	.endr

	## And save the floating point registers
	.set _freg, 0
	.rept MAX_SAVE_FREG+1
	stfd _freg, ex_ppcreg_freg+_freg*8(r1)
	.set _freg, _freg+1
	.endr

	## Save registers used in vsr (r14+r15)
	stw r14, ex_ppcreg_reg+14*4(r1)
	stw r15, ex_ppcreg_reg+15*4(r1)

	## get remaining cpu registers
	mfxer r3
	mfctr r5
	mfdar r6
	mfdsisr r7
	mfpvr r8
	mfsrr0 r9
	mfsrr1 r10

	## and save them
	stw r3, ex_ppcreg_xer(r1)
	stw r5, ex_ppcreg_ctr(r1)
	stw r6, ex_ppcreg_dar(r1)
	stw r7, ex_ppcreg_dsisr(r1)
	stw r8, ex_ppcreg_pvr(r1)
	stw r9, ex_ppcreg_pc(r1)
	stw r10, ex_ppcreg_msr(r1)

	## The entire CPU state is now stashed on the stack. Call into C
	## to do something with it
	mr r3, sp				## r3 = register dump
	subi sp, sp, ppc_stack_frame_size	## Make a NULL frame
	li r0, 0				## r0 = 0
	stw r0, 0(sp)				## backchain = 0
	stw r0, 8(sp)				## return pc = 0

	stwu sp, -ppc_stack_frame_size(sp)	## Create new stack frame where C can
						## save the LR
	lwi r5, restore_state			## store the return address in the LR
	mtlr r5
		
	.extern cyg_hal_exception_handler
	b cyg_hal_exception_handler		## call C-code. Upon return it will pass
						## control to restore_state

	##-----------------------------------------------------------------
	## Common interrupt handling code
	## Decoding the interrupt 
	## Important NOTICE : CYG_ISR_TABLE_SIZE must match CYG_ISR_COUNT defined
	## in hal_intr.h

#define CYG_ISR_TABLE_SIZE 136
	
	.macro decode_interrupt dreg, state
	lwz \dreg, ex_ppcreg_vector(\state)	## retrieve the vector number
	rlwinm. \dreg, \dreg, 22, 31, 31	## isolate bit 21
	beq 0f					## done if decrementer (vec 0)
	lwi \dreg, CYGARC_REG_IMM_SIVEC		## if external, get SIU vector
	lbz \dreg, 0(\dreg)
	srwi \dreg, \dreg, 2
	addi \dreg, \dreg, 1			## skip decrementer vector
0:	stw \dreg, ex_ppcreg_vector(\state)	## update the vector in state frame
	slwi \dreg, \dreg, 2			## convert to byte offset
	.endm

__default_interrupt_vsr:
	## We come here with all registers containing their pre-exception
	## values, except :
	## R3	 = least significant bits of of vector address
	## R4	 = saved CR
	## R5	 = saved LR
	## LR	 = VSR address
	## SPRG1 = old R3
	## SPRG2 = old R4
	## SPRG3 = old R5
	## SRR0	 = old PC
	## SRR1	 = old MSR
	subi r1, r1, ppc_exception_decrement
						## Leave space for registers and a safety
						## margin
	stw r3, ex_ppcreg_vector(r1)		## stash vector
	stw r4, ex_ppcreg_cr(r1)		## stash CR
	stw r5, ex_ppcreg_lr(r1)		## stash LR

	## enable MMU
	lwi r3, CYG_MSR
	sync
	mtmsr r3
	sync

	mfspr r3, sprg1				## save original r3
	stw r3, ex_ppcreg_reg+3*4(r1)
	mfspr r4, sprg2				## save original r4
	stw r4, ex_ppcreg_reg+4*4(r1)
	mfspr r5, sprg3				## save original r5
	stw r5,	ex_ppcreg_reg+5*4(r1)	

	stw r0, ex_ppcreg_reg+0*4(r1)		## save r0
	stw r2, ex_ppcreg_reg+2*4(r1)		## save r2

	mr r3, r1				## recreate original r1 and
	addi r3, r3, ppc_exception_decrement	## store it in state
	stw r3, ex_ppcreg_reg+1*4(r1)

	## Now save all other integer registers
	.set _reg, 6
	.rept MAX_SAVE_REG+1-6
	stw _reg, ex_ppcreg_reg+_reg*4(r1)
	.set _reg, _reg+1
	.endr

	## And save the floating point registers
	.set _freg, 0
	.rept MAX_SAVE_FREG+1
	stfd _freg, ex_ppcreg_freg+_freg*8(r1)
	.set _freg, _freg+1
	.endr

	## Save registers used in vsr (r14+r15)
	stw r14, ex_ppcreg_reg+14*4(r1)
	stw r15, ex_ppcreg_reg+15*4(r1)

	## Get the remaining CPU registers
	mfxer r3
	mfctr r5
	mfsrr0 r6
	mfsrr1 r7

	## and store them
	stw r3, ex_ppcreg_xer(r1)
	stw r5, ex_ppcreg_ctr(r1)
	stw r6, ex_ppcreg_pc(r1)
	stw r7, ex_ppcreg_msr(r1)

	## OK, the entire CPU state is now stashed on the stack.
	## Increment the scheduler lock and and call the ISR for this vector
#ifdef CYGFUN_HAL_COMMON_KERNEL_SUPPORT
	.extern cyg_scheduler_sched_lock
	lwi r3, cyg_scheduler_sched_lock
	lwz r4, 0(r3)
	addi r4, r4, 1
	stw r4, 0(r3)
#endif

	mr r14, sp				## r14 = register dump

#ifdef CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK
	lwi r3, __interrupt_stack		## stack top
	lwi r4, __interrupt_stack_base		## stack base
	sub. r5, sp, r4				## sp - base
	blt 1f					## if < 0 - not on istack
	sub. r5, r3, sp				## top - sp
	bgt 2f					## if > 0 - already on istack

1:	mr sp, r3				## switch to istack
2:	stwu r14, -4(sp)			## save old SP on stack
#endif

	subi sp, sp, ppc_stack_frame_size	## make a null frame

	li r0, 0				## r0 = 0
	stw r0, 0(sp)				## backchain = 0
	stw r0, 8(sp)				## return pc = 0

	stwu sp, -ppc_stack_frame_size(sp)	## Make a new stack frame where C-code
						## can save LR
#if defined(CYGPKG_KERNEL_INSTRUMENT) && defined(CYGDBG_KERNEL_INSTRUMENT_INTR)
	lwi r3, 0x0301				## r3 = type = INTR, RAISE
	lwz r4, ex_ppcreg_vector(r14)		## arg1 = vector address
	srwi r4, r4, 8				## arg1 = vector number
	xor r5, r5, r5				## arg2 = 0
	bl cyg_instrument			## call instrument function
#endif
		
	decode_interrupt r15, r14		## get table index

#ifdef CYGDBG_HAL_DEBUG_GDB_BREAK_SUPPORT
        ## IMPORTANT NOTICE
        ## The stub has break support
	## Since the MPC555 micro allows for very versatile interrupt
	## configuration, it is not advisable to allow the stub to configure the
	## interrupts on the MPC555. This should be done by a more complete build.
	## The break support will only work if the RX interrupt for the debug port
	## occurs on the lowest SIU priority level (i.e. level 7)
	## See file "mpc555_isrsys.cxx" for an example of how to setup interrupt
	## priorities.
        lwi r6, hal_interrupt_handlers		## load the special isr for break support
        lwi r3, CYG_ISR_TABLE_SIZE
        slwi r3, r3, 2
        lwzx r6, r6, r3
	
	lwz r3, ex_ppcreg_pc(r14)		## pass the program counter as an argument
        mtctr r6
        bctrl

	cmpwi r3, 0x0000			## Call isr proper, interrupt was not meant for GDB stub
	beq 2f					## R3 is 0 when skipping to avoid DSR call
1:
#endif

	lwz r3, ex_ppcreg_vector(r14)		## retrieve decoded vector #

	lwi r6, hal_interrupt_handlers		## get interrupt handler table
	lwzx r6, r6, r15			## load routine pointer

	lwi r4, hal_interrupt_data		## get interrupt data table
	lwzx r4, r4, r15			## load data pointer
						## r4 = data atgument
	mtctr r6				## put isr in ctr
	bctrl					## branch to ctr register and link

#ifdef CYGDBG_HAL_DEBUG_GDB_BREAK_SUPPORT
2:	## if interrupt was caused by GDB, the ISR above is skipped by jumping here
#endif

#ifdef CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK
	## if we are returning from the last nested interrupt, move back to the thread stack.
	## interrupt_end() must be called on the thread stack since it potentially causes
	## a context switch. Since we have provided for the top of the stack location to contain 
	## the sp we need to go back there, just pop it off and put it in SP

	lwz sp, ppc_stack_frame_size*2(sp)	## sp = *sp
	subi sp, sp, ppc_stack_frame_size	## make a NULL frame

	li r0, 0				## r0 = 0
	stw r0, 0(sp)				## backchain = 0
	stw r0, 8(sp)				## return pc = 0

	stwu sp, -ppc_stack_frame_size(sp)	## create a new stack frame where C-code can save 
						## LR
#endif

#ifdef CYGFUN_HAL_COMMON_KERNEL_SUPPORT
	## We only need to call _interrupt_end() when there is a kernel present to do any
	## tidying up.

	## Only r3 bit 11 will indicate whether a DSR is to be posted. Pass this together
	## with with a pointer to the interrupt object we have just used to the interrupt tidy
	## up routine.

	## Note that r14 and r15 are defined to be preserved accross calls by the calling convention,
	## so they still contain the register dump and the vector number respectively.
	## Note: the instructions restoring the msr mask out the POW bit. When eCos gets powermanagement
	## aware, this bit will have to be set to the correct state, rather than just cleared.

	lwi r4, hal_interrupt_objects		## get interrupt object table
	lwzx r4, r4, r15			## load opbject pointer

	mr r5, r14				## arg3 = saved register dump

	lwz r6, ex_ppcreg_msr(r14)		## restore msr, including
	lis r5, 0xfffe				## interrupt enable
	andc r6, r6, r5				## mask out reserved bits
	sync					## (and POW !)
	mtmsr r6
	sync
	
        .extern interrupt_end
	bl interrupt_end			## call into C to finish off
#endif

	##-----------------------------------------------------------------
	## Exception is completely handled. Restore the CPU state and
	## continue normal work
restore_state:
	addi sp, sp, ppc_stack_frame_size*2	## retrieve the CPU state

	## get sprs we want to restore
	lwz r3, ex_ppcreg_cr(sp)
	lwz r4, ex_ppcreg_pc(sp)
	lwz r5, ex_ppcreg_msr(sp)
	lwz r6, ex_ppcreg_xer(sp)
	lwz r7, ex_ppcreg_lr(sp)
	lwz r8, ex_ppcreg_ctr(sp)

	## stuff some of them into the CPU
	mtxer r6
	mtlr r7
	mtctr r8

	## restore r0 and r2
	lwz r0, ex_ppcreg_reg+0*4(sp)
	lwz r2, ex_ppcreg_reg+2*4(sp)

	## restore registers used in vsr (r14+r15)
	lwz r14, ex_ppcreg_reg+14*4(r1)
	lwz r15, ex_ppcreg_reg+15*4(r1)

	## Now restore the integer registers
	.set _reg, 6
	.rept MAX_SAVE_REG+1-6
	lwz _reg, ex_ppcreg_reg+_reg*4(r1)
	.set _reg, _reg+1
	.endr

	## and restore the floating point registers
	.set _freg, 0
	.rept MAX_SAVE_FREG+1
	lfd _freg, ex_ppcreg_freg+_freg*8(r1)
	.set _freg, _freg+1
	.endr

	## Ok, all registers loaded except
	## r1 = ppcregs
	## r3 = ccr
	## r4 = srr0 = pc
	## r5 = srr1 = msr
	## Loading srr0 and srr1 has to be done under disable interrupt, a new
	## interrupt would destroy them
	mtcr r3

	CLI

	mtsrr0 r4				## load old pc
	mtsrr1 r5				## load old msr

	## And finally restore the last registers
	lwz r3, ex_ppcreg_reg+3*4(r1)		## load r3 value
	lwz r4, ex_ppcreg_reg+4*4(r1)		## load r4 value
	lwz r5, ex_ppcreg_reg+5*4(r1)		## load r5 value
	lwz r1, ex_ppcreg_reg+1*4(r1) 		## restore r1

	sync					## settle things down
	isync
	rfi

	##-----------------------------------------------------------------
	## This table contains a VSR pointer for each defined exception.
	## All of these except 5 and 9 point to the __default_exception_vsr
	## above, 5 and 9 point to the __default_interrupt_vsr. Entries in
	## this table may be changed using Cyg_Interrupt::set_vsr().
        .section ".exc_vsr","ax" 
	.global hal_vsr_table

hal_vsr_table:
	.long __default_exception_vsr	## 0 reserved
	.long __default_exception_vsr	## 1 system reset
	.long __default_exception_vsr	## 2 machine check
	.long __default_exception_vsr	## 3 reserved
	.long __default_exception_vsr	## 4 reserved

	.long __default_interrupt_vsr	## 5 external interrupt

	.long __default_exception_vsr	## 6 alignment
	.long __default_exception_vsr	## 7 program
	.long __default_exception_vsr	## 8 floating point unavailable

	.long __default_interrupt_vsr	## 9 decrementer

	.long __default_exception_vsr   ## 10 reserved 0x00a00
	.long __default_exception_vsr   ## 11 reserved 0x00b00
	.long __default_exception_vsr	## 12 system call
	.long __default_exception_vsr   ## 13 trace
	.long __default_exception_vsr	## 14 floating point assist
	.long __default_exception_vsr	## 15 reserved 0x0f00

	.long __default_exception_vsr	## 16 software emulation
	.long __default_exception_vsr	## 17 reserved 0x01100
	.long __default_exception_vsr	## 18 reserved 0x01200
	.long __default_exception_vsr	## 19 instruction protection error
	.long __default_exception_vsr	## 20 data protection error
	.long __default_exception_vsr	## 21 reserved 0x01500
	.long __default_exception_vsr	## 22 reserved 0x01600
	.long __default_exception_vsr	## 23 reserved 0x01700
	.long __default_exception_vsr	## 24 reserved 0x01800
	.long __default_exception_vsr	## 25 reserved 0x01900
	.long __default_exception_vsr	## 26 reserved 0x0a000
	.long __default_exception_vsr	## 27 reserved 0x01b00
	.long __default_exception_vsr	## 28 data breakpoint
	.long __default_exception_vsr	## 29 instruction breakpoint
	.long __default_exception_vsr	## 30 peripheral breakpoint
	.long __default_exception_vsr	## 31 non masakable external breakpoint

	##--------------------------------------------------------------------
	## Interrupt vector tables
	## These tables contain the isr, data and object pointers used to deliver
	## interrupts to user code
        .section ".isr_handlers","ax"
	.extern hal_default_isr
	.global hal_interrupt_handlers
hal_interrupt_handlers:
	.rept CYG_ISR_TABLE_SIZE
	.long hal_default_isr
	.endr
#if defined(CYGDBG_HAL_DEBUG_GDB_BREAK_SUPPORT)
#if defined(CYG_HAL_STARTUP_STUBS) || defined(CYG_HAL_STARTUP_ROM)
	.extern cyg_hal_gdb_isr
	.long cyg_hal_gdb_isr           ## Make this special vector accesible to non stub code
#else
	.long 0                         ## This is a noload section for RAM startup, make sure the address is OK
#endif					## contents is of no importance.
#endif

	.section ".isr_data","ax"
	.global hal_interrupt_data
hal_interrupt_data:
	.rept CYG_ISR_TABLE_SIZE
	.long 0
	.endr

	.section ".isr_obj","ax"
	.global hal_interrupt_objects
hal_interrupt_objects:
	.rept CYG_ISR_TABLE_SIZE
	.long 0
	.endr
		
	##--------------------------------------------------------------------
	## Allocate the interrupt stack
	.section ".isrstack"
	.global __interrupt_stack
	.balign 16
__interrupt_stack_base:
	.rept CYGNUM_HAL_COMMON_INTERRUPTS_STACK_SIZE
	.byte 0
	.endr
__interrupt_stack:
	.balign 16
	.long 0,0,0,0,0,0,0,0
