#ifndef CAN_H
#define CAN_H

//----------------------------------
// Includes and forward declarations
//----------------------------------
#include <cyg/io/canio.h>
#include <cyg/io/devtab.h>

//-----------------
// Type definitions
//-----------------
typedef struct can_channel can_channel;
typedef struct can_funs    can_funs;
typedef struct can_message can_message;

//-------------------------------------------------------------------------------------------------
// Callback functions to allow the low level device driver for accessing higher layer functionality
//-------------------------------------------------------------------------------------------------
typedef struct {
  void (* can_init)(can_channel * a_chan, bool a_force);
  void (* can_xmt_message)(can_channel * a_chan);
  void (* can_rcv_message)(can_channel * a_chan, const can_message * a_message);
}can_callbacks_t;

#define CAN_CALLBACKS(_l, _init, _xmt, _rcv)                          \
can_callbacks_t _l = {                                                \
  _init,                                                              \
  _xmt,                                                               \
  _rcv                                                                \
}

extern can_callbacks_t can_callbacks;

//---------------------------------------
// The definition of a CAN message buffer
//---------------------------------------
typedef struct {
  can_message *         m_data;
  volatile int          m_put;
  volatile int          m_get;
  int                   m_len;
  int                   m_low_water;
  cyg_drv_cond_t        m_wait;
  cyg_drv_mutex_t       m_lock;
  bool                  m_waiting;
  bool                  m_abort;
} can_buf_t;

#define CAN_BUF_INIT(_data, _len)  \
  {_data, 0, 0, _len}

//----------------------------------------
// Private data that describe this channel
//----------------------------------------
struct can_channel {
  can_funs *             m_funs;
  can_callbacks_t *      m_callbacks;
  void *                 m_dev_priv;                 // Whatever is needed by actual device routines
  can_info_t             m_config;
  bool                   m_init;
  can_buf_t              m_out_buf;
  can_buf_t              m_in_buf;
  can_rxid_action_t      m_action1;
  can_rxid_action_t      m_action2;
  bool                   m_transmitting;
  Cyg_ErrNo              m_error; 
  cyg_uint8              m_id_filter[256];
  cyg_uint16             m_rtr_id[16];
};

#define CAN_CHANNEL(_l,                                               \
                    _funs,                                            \
                    _dev_priv,                                        \
                    _baud,                                            \
                    _rx_mode,                                         \
                    _tx_mode,                                         \
                    _out_buf, _out_buf_len,                           \
                    _in_buf, _in_buf_len)                             \
can_channel _l = {                                                    \
  &_funs,                                                             \
  &can_callbacks,                                                     \
  &(_dev_priv),                                                       \
  CAN_INFO_INIT(_baud, _rx_mode, _tx_mode),                           \
  false,                                                              \
  CAN_BUF_INIT(_out_buf, _out_buf_len),                               \
  CAN_BUF_INIT(_in_buf, _in_buf_len),                                 \
  CAN_RXID_ACTION_INIT(0, 0),                                         \
  CAN_RXID_ACTION_INIT(0, 0),                                         \
  false                                                               \
}

//----------------------------
// Definition of a can message
//----------------------------
struct can_message {
  cyg_uint16 m_CAN_id;
  cyg_uint16 m_RTR;
  cyg_uint16 m_DLC;
  cyg_uint8  m_Data[8];
};

//------------------------------
// Low level interface functions
//------------------------------
struct can_funs {
  bool (* set_config)(can_channel * a_chan, can_info_t * a_config);                  // Set baudrate and electrical characteristics of the bus
  bool (* xmit_message)(can_channel * a_chan, const can_message * a_msg);            // Transmit the message
  void (* start_xmit)(can_channel * a_chan);                                         // Start the transmit process
  bool (* set_rxid1_action)(can_channel * a_chan, can_rxid_action_t * a_rxidaction); // Assign a callback function
  bool (* set_rxid2_action)(can_channel * a_chan, can_rxid_action_t * a_rxidaction); // Assign a callback function
  bool (* add_rx_id)(can_channel * a_chan, cyg_uint16 a_id);                         // Add the id to the message filter
  bool (* remove_rx_id)(can_channel * a_chan, cyg_uint16 a_id);                      // Remove the id from the message filter
  bool (* add_rtr_id)(can_channel * a_chan, cyg_uint16 a_id);                        // Add the id to the RTR filter
  bool (* remove_rtr_id)(can_channel * a_chan, cyg_uint16 a_id);                     // Remove the id from the RTR filter
  bool (* update_rtr_buffer)(can_channel * a_chan, const can_message * a_msg);       // Update the contents of a RTR buffer
  bool (* reset)(can_channel * a_chan);
};

#define CAN_FUNS(_l,                                                  \
                 _set_config,                                         \
                 _xmit_message,                                       \
                 _start_xmit,                                         \
                 _action1,                                            \
                 _action2,                                            \
                 _add_rx_id,                                          \
                 _remove_rx_id,                                       \
                 _add_rtr_id,                                         \
                 _remove_rtr_id,                                      \
                 _update_rtr_buffer,                                  \
                 _reset)                                              \
  can_funs _l = {                                                     \
    _set_config,                                                      \
    _xmit_message,                                                    \
    _start_xmit,                                                      \
    _action1,                                                         \
    _action2,                                                         \
    _add_rx_id,                                                       \
    _remove_rx_id,                                                    \
    _add_rtr_id,                                                      \
    _remove_rtr_id,                                                   \
    _update_rtr_buffer,                                               \
    _reset                                                            \
  };

//---------------------------------
// Define the device driver entries
//---------------------------------
extern cyg_devio_table_t can_devio;

#endif // CAN_H
